/*-----------------------------------------------------------------
  usbclock by Jack Botner

  interrupt.c

  Contains all the interrupt handlers for the program.

  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include "common.h"

// Global variables

extern volatile unsigned char ucRxBuf[RX_BUF_SIZE];
extern volatile uint8_t ucRxBytes;
extern volatile unsigned char ucTxBuf[TX_BUF_SIZE];
extern volatile uint8_t ucTxBytes, ucTxIndex;
extern volatile uint8_t ucDiagCode, ucDiagData;

extern volatile uint8_t  ucEventFlags;
extern volatile uint8_t  uc100msCtr;
extern volatile uint16_t uc100msCtr2;
extern volatile uint8_t  ucSynchronized;

// internal functions
static unsigned char checksum( volatile unsigned char *pucBuffer, int nBytes );

/*-----------------------------------------------------------------
  USART_RXC_vect - Rx complete USART interrupt
-----------------------------------------------------------------*/
ISR(USART_RX_vect)
{
  uint8_t	rxChar, ucCk, nn;

  if ( test_bit( UCSR0A, FE0 ) || test_bit( UCSR0A, DOR0 ) )
  {
    rxChar = UDR0;		// empty the data register

    set_bit( PORTC, PC4 );		// turn beeper on
    uc100msCtr = 1;				// for 100ms
    return;
  }
	
  rxChar = UDR0;		// empty the data register

  if ( rxChar == 0xee )
  {
	set_bit( ucEventFlags, EVENT_RX_SYNCHRONIZE );
    return;
  }
  
  if ( ucRxBytes >= RX_BUF_SIZE )
  {
	ucDiagCode = 1;
	ucDiagData = ucRxBytes;
	set_bit( ucEventFlags, EVENT_RX_FRAME_ERROR );
  }
  else
  {
    ucRxBuf[ucRxBytes] = rxChar;
	++ucRxBytes;

    if ( ucRxBytes > 2 )
	{
	  // Check for a complete frame
	  if ( ucRxBytes == ucRxBuf[1] )
	  {
	    // Verify the checksum
		nn = ucRxBytes - 1;
		//for ( ii=0, ucCk=0; ii<nn; ++ii )
		//  ucCk += ucRxBuf[ii];
		ucCk = checksum( ucRxBuf, nn );

		if ( ucCk == ucRxBuf[nn] )
		  set_bit( ucEventFlags, EVENT_RX_FRAME_COMPLETE );
		else
		{
	      ucDiagCode = 2;
	      ucDiagData = ucRxBuf[nn];
		  set_bit( ucEventFlags, EVENT_RX_FRAME_ERROR );
		}
	  }
	}
  }
}

/*-------------------------------------------------------------------
  Compute the checksum

  Returns: checksum
-------------------------------------------------------------------*/
static unsigned char checksum( volatile unsigned char *pucBuffer, int nBytes )
{
  int			ii;
  unsigned char ucCk=0;

  for ( ii=0; ii<nBytes; ++ii )
    ucCk += pucBuffer[ii];

  return ( ucCk & 0x7f );
}

/*-----------------------------------------------------------------
  USART_UDRE_vect - Tx data register empty USART interrupt
-----------------------------------------------------------------*/
ISR(USART_UDRE_vect)
{
  if ( ucTxIndex < ucTxBytes )
  {
	UDR0 = ucTxBuf[ucTxIndex];
	++ucTxIndex;
	if ( ucTxIndex == ucTxBytes )
	{
	  ucTxBytes = ucTxIndex = 0;
	  clear_bit( UCSR0B, UDRIE0 );		// disable further Tx interrupts
	}
  }
}

/*-----------------------------------------------------------------
  Timer 0 overflow interrupt handler - button debouncing

  To achieve an interval of 4 ms for sampling the pushbuttons.
  The clock period is 125 ns. Using a prescaler of 256 and 
  125 counts gives us an interrupt every 4 ms.
-----------------------------------------------------------------*/
ISR(TIMER0_OVF_vect)
{
  static uint8_t ucMask=1;
  static uint8_t ucButton1=0, ucButton1Prev=0;
  static uint8_t ucButton2=0, ucButton2Prev=0;

  // Read current input states

  if ( test_bit( PINC, 5 ) )		// switch off
    ucButton1 &= ~ucMask;
  else								// switch on
    ucButton1 |= ucMask;

  if ( test_bit( PINB, 0 ) )		// switch off
    ucButton2 &= ~ucMask;
  else								// switch on
    ucButton2 |= ucMask;

  ucMask = ucMask << 1;
  if ( !ucMask ) ucMask = 1;

  // Create button release events

  if ( ucButton1 == 0xff )
  {
    // The hold switch is closed now
	if ( !ucButton1Prev )
	  // The hold switch was open before
	  ucButton1Prev = ucButton1;		// remember the button has just been pressed
  }
  else
    if ( !ucButton1 )
	{
	  // The hold switch is open now
	  if ( ucButton1Prev == 0xff )
	  {
	    // The hold switch was closed before
		ucButton1Prev = ucButton1;	// remember the button has just been released
		set_bit( ucEventFlags, EVENT_BUTTON1_RELEASED );
	  }
	}

  if ( ucButton2 == 0xff )
  {
    // The clear switch is closed now
	if ( !ucButton2Prev )
	  // The clear switch was open before
	  ucButton2Prev = ucButton2;		// remember the button has just been pressed
  }
  else
    if ( !ucButton2 )
	{
	  // The clear switch is open now
	  if ( ucButton2Prev == 0xff )
	  {
	    // The clear switch was closed before
		ucButton2Prev = ucButton2;	// remember the button has just been pressed
		set_bit( ucEventFlags, EVENT_BUTTON2_RELEASED );
	  }
	}

  TCNT0 = INIT_CTR0;				// reinitialize counter
}

/*-----------------------------------------------------------------
  Timer 1 overflow interrupt handler - initiates adc conversion

  To achieve an interval of 100 ms for timing things. The clock
  period is 125ns. Using a prescaler of 256 and 3125 counts
  gives us an interrupt every 100 ms.
-----------------------------------------------------------------*/
ISR(TIMER1_OVF_vect)
{
  static uint8_t  uc100msCtr3=0;
  
  if ( uc100msCtr )			// a timing event is in progress
  {
    --uc100msCtr;
	if ( !uc100msCtr )
	  set_bit( ucEventFlags, EVENT_INTERVAL_EXPIRED );
  }

  if ( ucSynchronized )
  {
    if ( uc100msCtr2 )		// the system timeout
    {
	  --uc100msCtr2;
	  if ( !uc100msCtr2 )
	    set_bit( ucEventFlags, EVENT_SYSTEM_TIMEOUT );
    }
  }

  ++uc100msCtr3;
  if ( uc100msCtr3 == 3 )
  {
	if ( test_bit( PINC, PC1 ) )
	  clear_bit( PORTC, PC1 );        // turn off yellow diagnostic LED
	else
	  set_bit( PORTC, PC1 );          // turn on yellow diagnostic LED
	  
	uc100msCtr3 = 0;  
  }
  
  TCNT1 = INIT_CTR1;		// reinitialize counter
}
